package be.iminds.ilabt.jfed.testing.base;

import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.TestbedInfoSource;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticAggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.RspecFactory;
import be.iminds.ilabt.jfed.rspec.model.RspecFactoryFactory;
import be.iminds.ilabt.jfed.rspec.model.impl.BasicRspecFactory;
import be.iminds.ilabt.jfed.testing.shared.AutomatedTestExternalExecutableLocations;
import be.iminds.ilabt.jfed.testing.shared.AutomatedTestExternalExecutableLocationsImpl;
import be.iminds.ilabt.jfed.util.TargetAuthority;
import com.google.inject.AbstractModule;
import com.google.inject.Injector;
import com.google.inject.Module;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.inject.Inject;
import javax.inject.Singleton;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Objects;

/**
 * ApiTestFactory will create ApiTests.
 *
 * Requirements for creating:
 *   - injector with needed modules. Typical:
 *      - LowLevelModule           (minimal requirement)
 *      - HighLevelModule          (for experimenter API)
 *      - SfaOnlyExperimentModule  (for experimenter API)
 *      Note that the injector must provide the correct:
 *        - GeniUserProvider
 *        - Logger
 *   - Test config string
 *   - Test class (Class or String)
 *   - TargetAuthority (if relevant for test, may be null for some tests)
 */
public class ApiTestFactory {
    private final static org.slf4j.Logger LOG = LoggerFactory.getLogger(ApiTestFactory.class);

    @Nonnull protected final Injector injector;

    @Inject
    public ApiTestFactory(@Nonnull Injector injector) {
        this.injector = injector;
    }

    /**
     * This constructor checks if the logger and geniUserProvider
     * are the same as provided by the injector.
     *
     * @param injector
     * @param logger
     * @param geniUserProvider
     */
    public ApiTestFactory(@Nonnull Injector injector,
                          @Nonnull Logger logger,
                          @Nonnull GeniUserProvider geniUserProvider) {
        this.injector = injector;
        Logger injectorLogger = injector.getInstance(Logger.class);
        GeniUserProvider injectorGeniUserProvider = injector.getInstance(GeniUserProvider.class);

        //noinspection ObjectEquality
        if (logger != injectorLogger)
            throw new IllegalArgumentException("Injector must provided same logger");

        //noinspection ObjectEquality
        if (geniUserProvider != injectorGeniUserProvider)
            throw new IllegalArgumentException("Injector must provided same geniUserProvider");
    }

    @Nonnull
    public ApiTest createApiTest(@Nonnull Logger logger,
                                 @Nonnull TargetAuthority testedAuthority,
                                 @Nonnull String testClassName,
                                 @Nonnull String testConfigString) throws InvalidApiTestException {
        try {
            Class<?> foundClass = Class.forName(testClassName);
            if (!ApiTest.class.isAssignableFrom(foundClass)) {
                throw new InvalidApiTestException("The test class \"" + testClassName + "\" is not an ApiTest");
            }
            @SuppressWarnings("unchecked")
            Class<? extends ApiTest> testClass = (Class<? extends ApiTest>) foundClass;
            return createApiTest(logger, testedAuthority, testClass, testConfigString);
        } catch (ClassNotFoundException e) {
            throw new InvalidApiTestException("The test class \"" + testClassName + "\" cannot be found", e);
        }
    }

    @Nonnull
    public ApiTest createApiTest(@Nonnull Logger logger,
                                 @Nonnull final TargetAuthority testedAuthority,
                                 @Nonnull Class<? extends ApiTest> testClass,
                                 @Nonnull String testConfigString) throws InvalidApiTestException {
        @Nonnull ApiTestMetaData apiTestMetaData = ApiTest.getMetaData(testClass);
        assert apiTestMetaData != null;
        @Nonnull Class/*<? extends ApiTestConfig>*/ testConfigClass = apiTestMetaData.getConfigClass();
        @Nonnull ApiTestConfig testConfigInstance = apiTestMetaData.parseApiTestConfig(testConfigString);
        assert testConfigClass != null;
        assert apiTestMetaData != null;

        //Note: all tests MUST have a constructor with @Inject

        boolean hasTargetBinding = hasBinding(TargetAuthority.class);
        boolean hasTestConfig = hasBinding(ApiTestConfig.class);

        if (hasTargetBinding) {
            TargetAuthority target = injector.getInstance(TargetAuthority.class);
            assert target != null;
            if (!Objects.equals(target.getOrigUrn(), testedAuthority.getOrigUrn()))
                throw new RuntimeException(
                        "Internal error: the injector was already bound for another authority! " +
                                "already bound to: " + testedAuthority.getOrigUrn() + " new target: " + target.getOrigUrn());
        }

        List<Module> extraModules = new ArrayList<>();

        if (!hasTargetBinding) {
            extraModules.add(new AbstractModule() {
                @Override
                protected void configure() {
                    bind(TargetAuthority.class).toInstance(testedAuthority);
                }
            });
        }
        if (!hasTestConfig) {
            extraModules.add(new AbstractModule() {
                @Override
                protected void configure() {
                    //noinspection unchecked  (because type erase makes this all madness. But we know it's ok.)
                    bind(testConfigClass).toInstance(testConfigInstance);
                }
            });
        }

        //rspecFactory might be needed
        extraModules.add(new AbstractModule() {
            @Override
            protected void configure() {
                bind(RspecFactory.class).toInstance(
                        RspecFactoryFactory.getRspecFactoryInstance(ModelRspecType.BASIC)
                );
            }
        });
        AutomatedTestExternalExecutableLocations automatedTestExternalExecutableLocations =
                new AutomatedTestExternalExecutableLocationsImpl();
        try {
            automatedTestExternalExecutableLocations.requireExe("ansible-playbook");
        } catch (AutomatedTestExternalExecutableLocations.ExecutableNotFoundException e) {
            //throw new RuntimeException("required exe not found", e);
            LOG.error("Required exe not found: test might fail later i fit is needed!", e);
        }
        extraModules.add(new AbstractModule() {
            @Override
            protected void configure() {
                bind(AutomatedTestExternalExecutableLocations.class)
                        .toInstance(automatedTestExternalExecutableLocations);
            }
        });

        Injector usedInjector;
        if (!extraModules.isEmpty()) {
            LOG.info("Constructing " + testClass.getName() + " using Guice child injector (was missing some things)");
            usedInjector = injector.createChildInjector(extraModules);
        } else {
            LOG.info("Constructing " + testClass.getName() + " using Guice injector (already has everything needed)");
            usedInjector = injector;
        }

        Logger injectorLogger = usedInjector.getInstance(Logger.class);
        //noinspection ObjectEquality
        if (logger != injectorLogger)
            throw new IllegalArgumentException("Injector must provided same logger");

        return usedInjector.getInstance(testClass);
    }

    private boolean hasBinding(Class clazz) {
        try {
            Object tst = injector.getInstance(clazz);
            return tst != null;
        } catch (com.google.inject.ConfigurationException e) {
            return false;
        }catch (Exception e) {
            LOG.warn("Injector did not already have target binding, and returned unexpected exception class.", e);
            return false;
        }
    }

    public class InvalidApiTestException extends Exception {
        public InvalidApiTestException() {
        }

        public InvalidApiTestException(String message) {
            super(message);
        }

        public InvalidApiTestException(String message, Throwable cause) {
            super(message, cause);
        }

        public InvalidApiTestException(Throwable cause) {
            super(cause);
        }

        public InvalidApiTestException(String message, Throwable cause, boolean enableSuppression, boolean writableStackTrace) {
            super(message, cause, enableSuppression, writableStackTrace);
        }
    }
}
